<!doctype html>{% load i18n static %}{# This is a very basic template that can be overwritten in projects #}
<html>
<head>
  <meta http-equiv="content-type" content="text/html; charset=utf-8">
  <title>{% block title %}{{ service_description.title }} {{ accept_operations|dictsort:0|join:"/" }}{% endblock %}</title>
  {% block link %}
      <link rel="stylesheet" type="text/css" href="{% static 'gisserver/bootstrap-5.3.6.min.css' %}">
      <link rel="stylesheet" type="text/css" href="{% static 'gisserver/index.css' %}">
  {% endblock %}
  {% block extrahead %}{% endblock %}
</head>
<body class="{% block body-class %}meta-page{% endblock %}">
  {% block content %}{% endblock %}
</body>
</html>
