{% load i18n %}
<h1>{{ service_description.title }} {{ accept_operations|dictsort:0|join:"/" }}</h1>
{% if service_description.abstract %}{{ service_description.abstract|linebreaks }}{% endif %}

{% if service_description.keywords or service_description.provider_name %}
<dl>
{% if service_description.keywords %}
<dt>{% translate "Keywords" %}:</dt><dd>{{ service_description.keywords|join:", " }}</dd>
{% endif %}
{% if service_description.provider_name %}
  <dt>{% translate "Provider" %}:</dt>
  <dd>{% if service_description.provider_site %}<a href="{{ service_description.provider_site }}">{% endif %}{{ service_description.provider_name }}{% if service_description.provider_site %}</a></dd>{% endif %}
  {% if service_description.contact_person %}<dt>{% translate "Contact" %}:</dt><dd>{{ service_description.contact_person }}</dd>{% endif %}
{% endif %}
</dl>
{% endif %}

{% if connect_url %}
<h2>{% translate "Using This WFS" %}</h2>
<p>{% translate "Add the following URL to your GIS application:" %}</p>
<p class="connect-url"><samp>{{ connect_url }}</samp></p>
{% endif %}
