
{% extends "gisserver/base.html" %}{% load i18n %}

{% block title %}{{ service_description.title }} {{ accept_operations|dictsort:0|join:"/" }}{% endblock %}
{% block content %}
  <header id="wfs-header">
    {% block header %}{% include "gisserver/service_description.html" %}{% endblock %}
  </header>

  <main id="wfs-main">
    {% block main %}
      {% if wfs_features %}
        <h2>{% translate "WFS Feature Types" %}</h2>
        {% for feature_type in wfs_features %}
          <article class="wfs-feature-type">{% include "gisserver/wfs/feature_type.html" %}</article>
        {% endfor %}
      {% endif %}
    {% endblock %}
  </main>
{% endblock %}
