/* Common page styling */

/*
  This general CSS that can be disabled
  by removing the .meta-page in the overwritten templates
*/

.meta-page {
  margin: 2rem 2rem 4rem 2rem;
}

.meta-page h2 { margin: 2rem 0 0.5rem 0; }
.meta-page h3 { margin: 1rem 0 0.5rem 0; }
.meta-page p { margin: 0 0 1rem 0; }

.meta-page code { color: #000; }

.meta-page dl {
  display: grid;
  grid-gap: 2px 16px;
  grid-template-columns: max-content;
}
.meta-page dd { grid-column-start: 2; margin: 0; }

.meta-page .wfs-feature-type { margin-bottom: 2.5rem; }
.meta-page .field-name { min-width: 20%; }
.meta-page .field-type { min-width: 25%; }

/* application-specific */
p.connect-url { margin-left: 2.5rem; font-weight: bold; }

tbody th { text-align: left; padding-right: 2rem; }
tbody td { padding-right: 2rem; }
.table > tbody > tr.complex-level-1 > th { padding-left: 20px; }
.table > tbody > tr.complex-level-2 > th { padding-left: 40px; }
.table > tbody > tr.complex-level-3 > th { padding-left: 60px; }
